# coding=utf-8
# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****

import os.path
import shutil
import tempfile
import unittest

from hashlib import sha1

from signing_clients.apps import (
    force_bytes,
    Manifest,
    JarExtractor,
    ParsingError,
    ZipFile,
    file_key,
    get_signature_serial_number,
    ignore_certain_metainf_files
)


MANIFEST_BODY = b"""Name: test-file
Digest-Algorithms: MD5 SHA1
MD5-Digest: 5BXJnAbD0DzWPCj6Ve/16w==
SHA1-Digest: 5Hwcbg1KaPMqjDAXV/XDq/f30U0=

Name: test-dir/nested-test-file
Digest-Algorithms: MD5 SHA1
MD5-Digest: 53dwfEn/GnFiWp0NQyqWlA==
SHA1-Digest: 4QzlrC8QyhQW1T0/Nay5kRr3gVo=
"""

MANIFEST = b"Manifest-Version: 1.0\n\n" + MANIFEST_BODY

SIGNATURE = b"""Signature-Version: 1.0
MD5-Digest-Manifest: dughN2Z8uP3eXIZm7GVpjA==
SHA1-Digest-Manifest: rnDwKcEuRYqy57DFyzwK/Luul+0=
"""

SIGNATURES_BODY = b"""Name: test-file
Digest-Algorithms: MD5 SHA1
MD5-Digest: jf86A0RSFH3oREWLkRAoIg==
SHA1-Digest: 9O+Do4sVlAh82x9ZYu1GbtyNToA=

Name: test-dir/nested-test-file
Digest-Algorithms: MD5 SHA1
MD5-Digest: YHTqD4SINsoZngWvbGIhAA==
SHA1-Digest: lys436ZGYKrHY6n57Iy/EyF5FuI=
"""

SIGNATURES = SIGNATURE + b"\n" + SIGNATURES_BODY

EXTRA_NEWLINE_SIGNATURE = b"""Signature-Version: 1.0
MD5-Digest-Manifest: A3IkNTcP2L6JzwQzkp+6Kg==
SHA1-Digest-Manifest: xQKf9C1JcIjfZoFxTWt3pzW2KYI=

"""

EXTRA_NEWLINE_SIGNATURES = EXTRA_NEWLINE_SIGNATURE + b"\n" + SIGNATURES_BODY + b"\n"  # noqa

CONTINUED_MANIFEST = MANIFEST + b"""
Name: test-dir/nested-test-dir/nested-test-dir/nested-test-dir/nested-te
 st-file
Digest-Algorithms: MD5 SHA1
MD5-Digest: 53dwfEn/GnFiWp0NQyqWlA==
SHA1-Digest: 4QzlrC8QyhQW1T0/Nay5kRr3gVo=
"""

# Test for 72 byte limit test
BROKEN_MANIFEST = MANIFEST + b"""
Name: test-dir/nested-test-dir/nested-test-dir/nested-test-dir/nested-test-file
Digest-Algorithms: MD5 SHA1
MD5-Digest: 53dwfEn/GnFiWp0NQyqWlA==
SHA1-Digest: 4QzlrC8QyhQW1T0/Nay5kRr3gVo=
"""

VERY_LONG_MANIFEST = b"""Manifest-Version: 1.0

Name: test-file
Digest-Algorithms: MD5 SHA1
MD5-Digest: 5BXJnAbD0DzWPCj6Ve/16w==
SHA1-Digest: 5Hwcbg1KaPMqjDAXV/XDq/f30U0=

Name: test-dir/nested-test-file
Digest-Algorithms: MD5 SHA1
MD5-Digest: 53dwfEn/GnFiWp0NQyqWlA==
SHA1-Digest: 4QzlrC8QyhQW1T0/Nay5kRr3gVo=

Name: test-dir/nested-test-dir-0/nested-test-dir-1/nested-test-dir-2/lon
 g-path-name-test
Digest-Algorithms: MD5 SHA1
MD5-Digest: 9bU/UEp83EbO/DWN3Ds/cg==
SHA1-Digest: lIbbwE8/2LFOD00+bJ/Wu80lR/I=
"""

# Test for Unicode
UNICODE_MANIFEST = u"""Manifest-Version: 1.0

Name: test-dir/súité-höñe.txt
Digest-Algorithms: MD5 SHA1
MD5-Digest: +ZqzWWcMtOrWxs8Xr/tt+A==
SHA1-Digest: B5HkCxgt6fXNr+dWPwXH2aALVWk=
"""


def get_file(fname):
    return os.path.join(os.path.dirname(__file__), fname)


class SigningTest(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix='tmp-signing-clients-test-')
        self.addCleanup(lambda: shutil.rmtree(self.tmpdir))

    def tmp_file(self, fname):
        return os.path.join(self.tmpdir, fname)

    def _extract(self, omit=False, newlines=False):
        return JarExtractor(get_file('test-jar.zip'),
                            omit_signature_sections=omit,
                            extra_newlines=newlines)

    def test_00_extractor(self):
        self.assertTrue(isinstance(self._extract(), JarExtractor))

    def test_01_manifest(self):
        extracted = self._extract()
        self.assertEqual(force_bytes(extracted.manifest), MANIFEST)
        extracted = self._extract(newlines=True)
        self.assertEqual(force_bytes(extracted.manifest), MANIFEST + b"\n")

    def test_02_signature(self):
        extracted = self._extract()
        self.assertEqual(force_bytes(extracted.signature), SIGNATURE)
        extracted = self._extract(newlines=True)
        self.assertEqual(force_bytes(extracted.signature), EXTRA_NEWLINE_SIGNATURE)

    def test_03_signatures(self):
        extracted = self._extract()
        self.assertEqual(force_bytes(extracted.signatures), SIGNATURES)
        extracted = self._extract(newlines=True)
        self.assertEqual(force_bytes(extracted.signatures), EXTRA_NEWLINE_SIGNATURES)

    def test_04_signatures_omit(self):
        extracted = self._extract(True)
        self.assertEqual(force_bytes(extracted.signatures), SIGNATURE)

    def test_05_continuation(self):
        manifest = Manifest.parse(CONTINUED_MANIFEST)
        self.assertEqual(force_bytes(manifest), CONTINUED_MANIFEST)

    def test_06_line_too_long(self):
        self.assertRaises(ParsingError, Manifest.parse, BROKEN_MANIFEST)

    def test_07_wrapping(self):
        extracted = JarExtractor(get_file('test-jar-long-path.zip'),
                                 omit_signature_sections=False)
        self.assertEqual(force_bytes(extracted.manifest), VERY_LONG_MANIFEST)

    def test_08_unicode(self):
        extracted = JarExtractor(get_file('test-jar-unicode.zip'),
                                 omit_signature_sections=False)
        self.assertEqual(
            force_bytes(extracted.manifest).decode('utf-8'), UNICODE_MANIFEST)

    def test_09_serial_number_extraction(self):
        with open(get_file('zigbert.test.pkcs7.der'), 'rb') as f:
            serialno = get_signature_serial_number(f.read())
        # Signature occured on Thursday, January 22nd 2015 at 11:02:22am PST
        # The signing service returns a Python time.time() value multiplied
        # by 1000 to get a (hopefully) truly unique serial number
        self.assertEqual(1421953342960, serialno)

    def test_10_resigning_manifest_exclusions(self):
        # This zip contains META-INF/manifest.mf, META-INF/zigbert.sf, and
        # META-INF/zigbert.rsa in addition to the contents of the basic test
        # archive test-jar.zip
        extracted = JarExtractor(get_file('test-jar-meta-inf-exclude.zip'),
                                 omit_signature_sections=True)
        self.assertEqual(force_bytes(extracted.manifest), MANIFEST)

    def test_11_make_signed(self):
        extracted = JarExtractor(get_file('test-jar.zip'),
                                 omit_signature_sections=True)
        # Not a valid signature but a PKCS7 data blob, at least
        with open(get_file('zigbert.test.pkcs7.der'), 'rb') as f:
            signature = f.read()
            signature_digest = sha1()
            signature_digest.update(signature)

        signed_file = self.tmp_file('test-jar-signed.zip')
        sigpath = 'zoidberg'
        extracted.make_signed(signature, signed_file, sigpath=sigpath)
        # Now verify the signed zipfile's contents
        with ZipFile(signed_file, 'r') as zin:
            # sorted(...) should result in the following order:
            files = ['test-file', 'META-INF/manifest.mf',
                     'META-INF/zoidberg.rsa',
                     'META-INF/zoidberg.sf',
                     'test-dir/', 'test-dir/nested-test-file']
            zfiles = [f.filename for f in sorted(zin.filelist, key=file_key)]
            self.assertEqual(files, zfiles)
            zip_sig_digest = sha1()
            zip_sig_digest.update(zin.read('META-INF/zoidberg.rsa'))

            self.assertEqual(signature_digest.hexdigest(),
                             zip_sig_digest.hexdigest())
        # And make sure the manifest is correct
        signed = JarExtractor(signed_file, omit_signature_sections=True)
        self.assertEqual(force_bytes(extracted.manifest), force_bytes(signed.manifest))

    def test_11_make_signed_default_sigpath(self):
        extracted = JarExtractor(get_file('test-jar.zip'),
                                 omit_signature_sections=True)
        # Not a valid signature but a PKCS7 data blob, at least
        with open(get_file('zigbert.test.pkcs7.der'), 'rb') as f:
            signature = f.read()
            signature_digest = sha1()
            signature_digest.update(signature)

        signed_file = self.tmp_file('test-jar-signed.zip')
        extracted.make_signed(signature, signed_file)

        with ZipFile(signed_file, 'r') as zin:
            files = ['test-file', 'META-INF/manifest.mf',
                     'META-INF/zigbert.rsa',
                     'META-INF/zigbert.sf',
                     'test-dir/', 'test-dir/nested-test-file']
            zfiles = [f.filename for f in sorted(zin.filelist, key=file_key)]
            self.assertEqual(files, zfiles)
            zip_sig_digest = sha1()
            zip_sig_digest.update(zin.read('META-INF/zigbert.rsa'))

            self.assertEqual(signature_digest.hexdigest(),
                             zip_sig_digest.hexdigest())

        signed = JarExtractor(signed_file, omit_signature_sections=True)
        self.assertEqual(force_bytes(extracted.manifest), force_bytes(signed.manifest))

    # See https://bugzil.la/1169574
    def test_12_metainf_case_sensitivity(self):
        self.assertTrue(ignore_certain_metainf_files('meta-inf/manifest.mf'))
        self.assertTrue(ignore_certain_metainf_files('MeTa-InF/MaNiFeSt.Mf'))
        self.assertFalse(ignore_certain_metainf_files('meta-inf/pickles.mf'))




# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****

import fnmatch
import hashlib
import itertools
import os.path
import re
from zipfile import ZipFile, ZIP_DEFLATED
from base64 import b64encode, b64decode

from six import PY3, string_types, text_type, python_2_unicode_compatible
from six.moves import cStringIO as StringIO

from asn1crypto import cms

# Lame hack to take advantage of a not well known OpenSSL flag.  This omits
# the S/MIME capabilities when generating a PKCS#7 signature.  If included,
# XPI signature verification breaks.
PKCS7_NOSMIMECAP = 0x200

headers_re = re.compile(
    r"""^((?:Manifest|Signature)-Version
          |Name
          |Digest-Algorithms
          |(?:MD5|SHA1)-Digest(?:-Manifest)?)
          \s*:\s*(.*)""", re.X | re.I)
continuation_re = re.compile(r"""^ (.*)""", re.I)
directory_re = re.compile(r"[\\/]$")


# partially copied from Django
def force_bytes(s, encoding='utf-8', errors='strict'):
    if isinstance(s, bytes):
        if encoding == 'utf-8':
            return s
        else:
            return s.decode('utf-8', errors).encode(encoding, errors)
    if not isinstance(s, string_types):
        try:
            if PY3:
                return text_type(s).encode(encoding)
            else:
                return bytes(s)
        except UnicodeEncodeError:
            return text_type(s).encode(encoding, errors)
    else:
        return s.encode(encoding, errors)


class ParsingError(Exception):
    pass


def ignore_certain_metainf_files(filename):
    """
    We do not support multiple signatures in XPI signing because the client
    side code makes some pretty reasonable assumptions about a single signature
    on any given JAR.  This function returns True if the file name given is one
    that we dispose of to prevent multiple signatures.
    """
    ignore = ("META-INF/manifest.mf",
              "META-INF/*.sf",
              "META-INF/*.rsa",
              "META-INF/*.dsa",
              "META-INF/ids.json")

    for glob in ignore:
        # Explicitly match against all upper case to prevent the kind of
        # runtime errors that lead to https://bugzil.la/1169574
        if fnmatch.fnmatchcase(filename.upper(), glob.upper()):
            return True
    return False


def file_key(zinfo):
    '''
    Sort keys for xpi files
    @param name: name of the file to generate the sort key from
    '''
    # Copied from xpisign.py's api.py and tweaked
    name = zinfo.filename
    prio = 4
    if name == 'install.rdf':
        prio = 1
    elif name in ["chrome.manifest", "icon.png", "icon64.png"]:
        prio = 2
    elif name in ["MPL", "GPL", "LGPL", "COPYING", "LICENSE", "license.txt"]:
        prio = 5
    parts = [prio] + list(os.path.split(name.lower()))
    return "%d-%s-%s" % tuple(parts)


def _digest(data):
    md5 = hashlib.md5()
    md5.update(force_bytes(data))
    sha1 = hashlib.sha1()
    sha1.update(force_bytes(data))
    return {'md5': md5.digest(), 'sha1': sha1.digest()}


@python_2_unicode_compatible
class Section(object):
    __slots__ = ('name', 'algos', 'digests')

    def __init__(self, name, algos=('md5', 'sha1'), digests={}):
        self.name = name
        self.algos = algos
        self.digests = digests

    def __str__(self):
        # Important thing to note: placement of newlines in these strings is
        # sensitive and should not be changed without reading through
        # http://docs.oracle.com/javase/7/docs/technotes/guides/jar/jar.html#JAR%20Manifest
        # thoroughly.
        algos = b''
        order = list(self.digests.keys())
        order.sort()
        for algo in order:
            algos += b' %s' % force_bytes(algo.upper())
        entry = b''
        # The spec for zip files only supports extended ASCII and UTF-8
        # See http://www.pkware.com/documents/casestudies/APPNOTE.TXT
        # and search for "language encoding" for details
        #
        # See https://bugzilla.mozilla.org/show_bug.cgi?id=1013347
        name = b'Name: %s' % force_bytes(self.name)

        # See https://bugzilla.mozilla.org/show_bug.cgi?id=841569#c35
        while name:
            entry += name[:72]
            name = name[72:]
            if name:
                entry += b'\n '
        entry += b'\n'
        entry += b'Digest-Algorithms:%s\n' % force_bytes(algos)
        for algo in order:
            entry += b'%s-Digest: %s\n' % (force_bytes(algo.upper()),
                                           b64encode(self.digests[algo]))
        return entry.decode('utf-8')


@python_2_unicode_compatible
class Manifest(list):
    version = '1.0'
    # Older versions of Firefox crash if a JAR manifest style file doesn't
    # end in a blank line("\n\n").  For more details see:
    # https://bugzilla.mozilla.org/show_bug.cgi?id=1158467
    extra_newline = False

    def __init__(self, *args, **kwargs):
        super(Manifest, self).__init__(*args)
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def parse(klass, buf):
        #version = None
        if hasattr(buf, 'readlines'):
            fest = buf
        else:
            fest = StringIO(force_bytes(buf).decode('utf-8'))
        kwargs = {}
        items = []
        item = {}
        header = ''  # persistent and used for accreting continuations
        lineno = 0
        # JAR spec requires two newlines at the end of a buffer to be parsed
        # and states that they should be appended if necessary.  Just throw
        # two newlines on every time because it won't hurt anything.
        for line in itertools.chain(fest.readlines(), "\n" * 2):
            lineno += 1
            line = line.rstrip()
            if len(line) > 72:
                raise ParsingError("Manifest parsing error: line too long "
                                   "(%d)" % lineno)
            # End of section
            if not line:
                if item:
                    items.append(Section(item.pop('name'), **item))
                    item = {}
                header = ''
                continue
            # continuation?
            continued = continuation_re.match(line)
            if continued:
                if not header:
                    raise ParsingError("Manifest parsing error: continued line"
                                       " without previous header! Line number"
                                       " %d" % lineno)
                item[header] += continued.group(1)
                continue
            match = headers_re.match(line)
            if not match:
                raise ParsingError("Unrecognized line format: \"%s\"" % line)
            header = match.group(1).lower()
            value = match.group(2)
            if '-version' == header[-8:]:
                # Not doing anything with these at the moment
                #payload = header[:-8]
                #version = value.strip()
                pass
            elif '-digest-manifest' == header[-16:]:
                if 'digest_manifests' not in kwargs:
                    kwargs['digest_manifests'] = {}
                kwargs['digest_manifests'][header[:-16]] = b64decode(value)
            elif 'name' == header:
                if directory_re.search(value):
                    continue
                item['name'] = value
                continue
            elif 'digest-algorithms' == header:
                item['algos'] = tuple(re.split('\s*', value.lower()))
                continue
            elif '-digest' == header[-7:]:
                if not 'digests' in item:
                    item['digests'] = {}
                item['digests'][header[:-7]] = b64decode(value)
                continue
        if len(kwargs):
            return klass(items, **kwargs)
        return klass(items)

    @property
    def header(self):
        return b"%s-Version: %s" % (
            force_bytes(type(self).__name__.title()),
            force_bytes(self.version))

    @property
    def body(self):
        return b"\n".join([force_bytes(i) for i in self])

    def __str__(self):
        segments = [self.header, b"", self.body]
        if self.extra_newline:
            segments.append(b"")
        return (b"\n".join(
            force_bytes(item) for item in segments)
        ).decode('utf-8')


@python_2_unicode_compatible
class Signature(Manifest):
    omit_individual_sections = True
    digest_manifests = {}
    filename = 'zigbert'

    @property
    def digest_manifest(self):
        return [b"%s-Digest-Manifest: %s" %
                (force_bytes(item[0].upper()), b64encode(item[1]))
                for item in sorted(self.digest_manifests.items())]

    @property
    def header(self):
        segments = [force_bytes(super(Signature, self).header)]
        segments.extend(self.digest_manifest)
        if self.extra_newline:
            segments.append(b"")
        return b"\n".join(force_bytes(item) for item in segments)

    # So we can omit the individual signature sections
    def __str__(self):
        if self.omit_individual_sections:
            return (self.header + b"\n").decode('utf-8')
        return super(Signature, self).__str__()


class JarExtractor(object):
    """
    Walks an archive, creating manifest.mf and signature.sf contents as it goes

    Can also generate a new signed archive, if given a PKCS#7 signature
    """

    def __init__(self, path, outpath=None, ids=None,
                 omit_signature_sections=False, extra_newlines=False):
        """
        """
        self.inpath = path
        self.outpath = outpath
        self._digests = []
        self.omit_sections = omit_signature_sections
        self.extra_newlines = extra_newlines
        self._manifest = None
        self._sig = None
        self.ids = ids

        def mksection(data, fname):
            digests = _digest(data)
            item = Section(fname, algos=tuple(digests.keys()),
                           digests=digests)
            self._digests.append(item)
        with ZipFile(self.inpath, 'r') as zin:
            for f in sorted(zin.filelist, key=file_key):
                # Skip directories and specific files found in META-INF/ that
                # are not permitted in the manifest
                if (directory_re.search(f.filename)
                        or ignore_certain_metainf_files(f.filename)):
                    continue
                mksection(zin.read(f.filename), f.filename)
            if ids:
                mksection(ids, 'META-INF/ids.json')

    def _sign(self, item):
        digests = _digest(str(item))
        return Section(item.name, algos=tuple(digests.keys()),
                       digests=digests)

    @property
    def manifest(self):
        if not self._manifest:
            self._manifest = Manifest(self._digests,
                                      extra_newline=self.extra_newlines)
        return self._manifest

    @property
    def signatures(self):
        # The META-INF/*.sf files should contain hashes of the individual
        # sections of the the META-INF/manifest.mf file.  So we generate those
        # signatures here
        if not self._sig:
            self._sig = Signature([self._sign(f) for f in self._digests],
                                  digest_manifests=_digest(force_bytes(self.manifest)),
                                  omit_individual_sections=self.omit_sections,
                                  extra_newline=self.extra_newlines)
        return self._sig

    @property
    def signature(self):
        # Returns only the x-Digest-Manifest signature and omits the individual
        # section signatures
        return self.signatures.header + b"\n"

    def make_signed(self, signature, outpath=None, sigpath=None):
        outpath = outpath or self.outpath
        if not outpath:
            raise IOError("No output file specified")

        if os.path.exists(outpath):
            raise IOError("File already exists: %s" % outpath)

        sigpath = sigpath or self.signatures.filename
        # Normalize to a simple filename with no extension or prefixed
        # directory
        sigpath = os.path.splitext(os.path.basename(sigpath))[0]
        sigpath = os.path.join('META-INF', sigpath)

        with ZipFile(self.inpath, 'r') as zin:
            with ZipFile(outpath, 'w', ZIP_DEFLATED) as zout:
                # The PKCS7 file("foo.rsa") *MUST* be the first file in the
                # archive to take advantage of Firefox's optimized downloading
                # of XPIs
                zout.writestr("%s.rsa" % sigpath, signature)
                for f in zin.infolist():
                    # Make sure we exclude any of our signature and manifest
                    # files
                    if ignore_certain_metainf_files(f.filename):
                        continue
                    zout.writestr(f, zin.read(f.filename))
                zout.writestr("META-INF/manifest.mf", str(self.manifest))
                zout.writestr("%s.sf" % sigpath, str(self.signatures))
                if self.ids is not None:
                    zout.writestr('META-INF/ids.json', self.ids)


# This is basically a dumbed down version of M2Crypto.SMIME.load_pkcs7 but
# that reads DER instead of only PEM formatted files
def get_signature_serial_number(pkcs7):
    """
    Extracts the serial number out of a DER formatted, detached PKCS7
    signature buffer
    """
    content = cms.ContentInfo.load(pkcs7)['content'].native

    # Fetch the certificate stack that is the list of signers
    # Since there should only be one in this use case, take the zeroth
    # cert in the stack and return its serial number
    try:
        return content['signer_infos'][0]['sid']['serial_number']
    except (IndexError, KeyError):
        return None



import sys
import os
import codecs
from setuptools import setup, find_packages

version = '1.1'

# Helper to publish to pypi. Just call `python setup.py publish`
if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    os.system('python setup.py bdist_wheel upload')
    print('You probably want to also tag the version now:')
    print('  git tag -a %s -m "version %s"' % (version, version))
    print('  git push --tags')
    sys.exit()


def read(*parts):
    filename = os.path.join(os.path.dirname(__file__), *parts)
    with codecs.open(filename, encoding='utf-8') as fp:
        return fp.read()


setup(
    name='signing_clients',
    version=version,
    description='Applications signature/manifest manipulator and receipt verifier',
    long_description=read('README.md'),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Security :: Cryptography',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    keywords='web services',
    author='Ryan Tilder',
    author_email='service-dev@mozilla.com',
    url='http://mozilla.org',
    install_requires=['asn1crypto', 'six'],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    test_suite='signing_clients.tests'
)
